"""Module containing all devices that are exposed via Carson Living"""

import logging
from abc import ABCMeta, abstractmethod

from carson_living.error import CarsonError

_LOGGER = logging.getLogger(__name__)


class _AbstractEntity(object):
    # pylint: disable=useless-object-inheritance
    """Updateable Base Entity

    Attributes:
        _update_callback:
            The callback that can be used to update the
            entity_payload:
        _entity_payload:
            The payload the the entity draws its internal
            state from.

    """
    __metaclass__ = ABCMeta

    def __init__(self, update_callback=None, entity_payload=None):
        self._update_callback = update_callback
        # Note, entity_payload is written in self.update()
        self._entity_payload = None

        # update internal representation
        self.update(entity_payload)

    def __repr__(self):
        return '{} with unique_entity_id {}'.format(
            self.__class__.__name__, self.unique_entity_id)

    @property
    @abstractmethod
    def entity_id(self):
        """Entity ID

        Returns:
            Domain-specific entity id.

        """
        raise NotImplementedError(
            'Derived Entity does not implement entity_id!')

    @property
    @abstractmethod
    def unique_entity_id(self):
        """Unqiue Entity ID

        Returns:
            Unique entity id across entire library.
        """
        raise NotImplementedError(
            'Derived Entity does not implement unique_entity_id!')

    @abstractmethod
    def _internal_update(self):
        """Update internal state

        Allows the entity to update internal state
        after _entity_payload changes
        """
        raise NotImplementedError(
            'Derived Entity does not implement _internal_update!')

    @property
    def entity_payload(self):
        """Entity Payload

        Returns: the raw entity_payload

        """
        return self._entity_payload

    def update(self, entity_payload=None):
        """Update the entity

        Updates the entity with a given entity_payload. This payload
        can be passed directly the update function OR if left empty,
        the entity updates via tha given update_callback

        Args:
            entity_payload: optional payload to setup the entity

        Raises:
            CarsonError:
                If neither a entity_payload or a update_callback
                was passes during initialization.

        """
        # If there is a entity_payload, use it over the callback.
        if entity_payload:
            self._entity_payload = entity_payload
            # Allow child class to perform internal updates
            self._internal_update()
            return

        if not self._update_callback:
            raise CarsonError(
                'Trying to update entity {} without external payload '
                'or a callback function'
                .format(self.unique_entity_id))

        _LOGGER.info(
            'Trying to updated entity %s from update callback',
            self.unique_entity_id)
        self._entity_payload = self._update_callback()

        # Allow child class to perform internal updates
        self._internal_update()


class _AbstractAPIEntity(_AbstractEntity):
    # pylint: disable=abstract-method
    __metaclass__ = ABCMeta

    def __init__(self, api, update_callback=None, entity_payload=None):
        self._api = api
        super(_AbstractAPIEntity, self).__init__(
            update_callback=update_callback,
            entity_payload=entity_payload)



"""Carson Living Entities"""

from carson_living.entities import (_AbstractEntity,
                                    _AbstractAPIEntity)

from carson_living.eagleeye import EagleEye

from carson_living.const import (C_API_URI,
                                 C_EEN_SESSION_ENDPOINT,
                                 C_DOOR_OPEN_ENDPOINT)

from carson_living.util import update_dictionary


class CarsonBuilding(_AbstractAPIEntity):
    """Carson Living Building Entity

    Attributes:
        _cameras:
            dict id->entity mapping of all camera entities
            associated with the given building.
        _doors:
            dict id->entity mapping of all door entities
            associated with the given building.
        _eagleeye:
            Eagle Eye API object that carries a building-specific
            authorization callback.


    """

    def __init__(self, api, entity_payload):
        self._cameras = {}
        self._doors = {}
        # Beware, entity building id must be injected early, since it is
        # required during object __init__
        self._eagleeye = EagleEye(
            lambda: self._get_eagleeye_session(api, entity_payload.get('id'))
        )

        super(CarsonBuilding, self).__init__(api,
                                             entity_payload=entity_payload)

    def __str__(self):
        pattern = """\
id: {entity_id}
name: {name}
number of cameras: {nr_cams}
number of door: {nr_doors}
number of units: {nr_units}
pmc: {pmc_name}"""
        return pattern.format(
            entity_id=self.entity_id,
            name=self.name,
            nr_cams=len(self.cameras),
            nr_doors=len(self.doors),
            nr_units=len(self.units),
            pmc_name=self.pmc_name
        )

    @staticmethod
    def _get_eagleeye_session(carson_api, building_id):
        """Retrieve a new eagle eye auth key and subdomain information

        Args:
            carson_api: The carson living api object
            building_id: The building id of the Carson property

        Returns:
            (tuple): tuple containing:

                sessionid(str): Eagle Eye authentication token for account
                subdomain(str): Eagle Eye subdomain to use with account

        """
        url = C_API_URI + C_EEN_SESSION_ENDPOINT.format(building_id)
        session = carson_api.authenticated_query(url)
        return session.get('sessionId'), session.get('activeBrandSubdomain')

    @property
    def entity_id(self):
        return self.entity_payload.get('id')

    @property
    def unique_entity_id(self):
        return 'carson_building_{}'.format(self.entity_id)

    def _internal_update(self):
        # Update Cameras from _entity_payload
        self._update_cameras()

        # Update Doors from _entity_payload
        self._update_doors()

    def _update_cameras(self):
        # Only Support Eagle_Eye right now
        # Update existing via Eagle Eye.
        self._eagleeye.update()

        # Cameras are managed by Eagle Eye API and
        # Carson Living only contains filter view of
        # Eagle Eye API
        self._cameras = {
            c['liveViewId']: self._eagleeye.get_camera(c['liveViewId'])
            for c in self.entity_payload.get('cameras')
            if c['provider'] == 'eagle_eye'
        }

    def _update_doors(self):
        update_doors = {d['id']: d for d in self.entity_payload.get('doors')}

        update_dictionary(
            self._doors,
            update_doors,
            lambda p: CarsonDoor(
                self._api,
                entity_payload=p))

    @property
    def eagleeye_api(self):
        """Eagle Eye API

        Returns: Eagle Eye API associated with that building.

        """
        return self._eagleeye

    @property
    def name(self):
        """Name

        Returns: The name of the building

        """
        return self.entity_payload.get('name')

    @property
    def type(self):
        """Type

        Returns: The type of building

        """
        return self.entity_payload.get('type')

    @property
    def cameras(self):
        """Cameras

        Returns: All camera entities associated with the building

        """
        return self._cameras.values()

    @property
    def doors(self):
        """Doors

        Returns: All door entities associated with the building

        """
        return self._doors.values()

    @property
    def payments_enabled(self):
        """Payments enabled

        Returns: True if that building has payments enabled

        """
        return self.entity_payload.get('paymentsEnabled')

    @property
    def area(self):
        """Area

        Returns:
            A dict with lat, lon, radius representing the geo
            location of the building

        """
        return self.entity_payload.get('area')

    @property
    def visitor_invite_enabled(self):
        """Visitor invite enabled

        Returns:
            True if the user is allowed to invite visitors
            to Carson Living

        """
        return self.entity_payload.get('visitorInviteEnabled')

    @property
    def doors_available(self):
        """Doors available

        Returns:
            True if the building has doors

        """
        return self.entity_payload.get('doorsAvailable')

    @property
    def pmc_name(self):
        """PMC

        Returns:
            Name of the property management company

        """
        return self.entity_payload.get('pmcName')

    @property
    def service_requests_enabled(self):
        """Service requests enabled

        Returns:
            True if the building allows to receive service requests

        """
        return self.entity_payload.get('serviceRequestsEnabled')

    @property
    def visitor_invites_left(self):
        """Visitor invites left

        Returns:
            Number of visitor invites that are left for the building
            (in regard to the current user)

        """
        return self.entity_payload.get('visitorInvitesLeft')

    @property
    def country(self):
        """Country

        Returns:
            Name of the country the building is in

        """
        return self.entity_payload.get('country')

    @property
    def state(self):
        """State

        Returns:
            Name of the state the building is in

        """
        return self.entity_payload.get('state')

    @property
    def timezone(self):
        """Timezone

        Returns:
            The tz database string the building is in

        """
        return self.entity_payload.get('timezone')

    @property
    def units(self):
        """List of units

        Returns:
            A list of units with element dict keys:
                name: Name of the Unit
                payments_enabled: True/False
        """
        return [
            {
                'name': u.get('name'),
                'payments_enabled': u.get('paymentsEnabled'),
            }
            for u in self.entity_payload.get('units')]


class CarsonUser(_AbstractEntity):
    """Carson Living User Entity

    """

    @property
    def unique_entity_id(self):
        return 'carson_user_{}'.format(self.entity_id)

    def _internal_update(self):
        pass

    @property
    def entity_id(self):
        return self.entity_payload.get('id')

    def __str__(self):
        pattern = """\
id: {entity_id}
first name: {first_name}
last name: {last_name}
contact info: {contact_info}
verified: {verfied}
is_admin: {is_admin}"""
        return pattern.format(
            entity_id=self.entity_id,
            first_name=self.first_name,
            last_name=self.last_name,
            contact_info=', '.join(
                ['{}: {}'.format(c['type'], c['contact_info'])
                 for c in self.contact_info]),
            verfied=self.verified,
            is_admin=self.is_admin
        )

    @property
    def first_name(self):
        """First Name

        Returns: First Name of the current User

        """
        return self.entity_payload.get('firstName')

    @property
    def last_name(self):
        """Last Name

        Returns: Last Name of the current User

        """
        return self.entity_payload.get('lastName')

    @property
    def contact_info(self):
        """List of contact information

        Returns:
            A list of contact information with element
            dict keys:
                contact_info: The contact info
                type: type of contact info
                primary: True/False
                verified: True/False
        """
        return [
            {
                'contact_info': c.get('contactInfo'),
                'type': c.get('type'),
                'primary': c.get('primary'),
                'verified': c.get('verified'),
            }
            for c in self.entity_payload.get('contactInfo')]

    @property
    def photo(self):
        """Photo

        Returns: Photo dict with keys:
            url: full-size photo
            thumbnail_url: thumbnail preview photo

        """
        photo = self.entity_payload.get('photo')
        return {
            'url': photo.get('url'),
            'thumbnail_url': photo.get('thumbnailUrl')
        }

    @property
    def verified(self):
        """Verified

        Returns: True if the user account is verified

        """
        return self.entity_payload.get('verified')

    @property
    def is_admin(self):
        """Verified

        Returns: True if the user account is an admin account

        """
        return self.entity_payload.get('isAdmin')

    @property
    def is_service(self):
        """Is Service

        Returns: True if the user account is a service account

        """
        return self.entity_payload.get('isService')


class CarsonDoor(_AbstractAPIEntity):
    """Carson Living Door Entity

    """

    @property
    def unique_entity_id(self):
        return 'carson_door_{}'.format(self.entity_id)

    def _internal_update(self):
        pass

    @property
    def entity_id(self):
        return self.entity_payload.get('id')

    def __str__(self):
        pattern = """\
id: {entity_id}
name: {name}
provider: {provider}
is_active: {is_active}
is_unit_door: {is_unit_door}
default_in_building: {default_in_building}"""
        return pattern.format(
            entity_id=self.entity_id,
            name=self.name,
            provider=self.provider,
            is_active=self.is_active,
            is_unit_door=self.is_unit_door,
            default_in_building=self.default_in_building
        )

    @property
    def name(self):
        """Name

        Returns: The name of the door

        """
        return self.entity_payload.get('name')

    @property
    def provider(self):
        """Provider

        Returns: The provider service of the door

        """
        return self.entity_payload.get('provider')

    @property
    def is_active(self):
        """Is active

        Returns: True if door is currently active

        """
        return self.entity_payload.get('isActive')

    @property
    def disabled(self):
        """disabled

        Returns: True if door is currently disabled

        """
        return self.entity_payload.get('disabled')

    @property
    def is_unit_door(self):
        """Is unit door

        Returns: True if door belongs to a unit

        """
        return self.entity_payload.get('isUnitDoor')

    @property
    def staff_only(self):
        """Staff only

        Returns: True if door can be used by staff only

        """
        return self.entity_payload.get('staffOnly')

    @property
    def default_in_building(self):
        """Default in building

        Returns: True if door is a default entrance door

        """
        return self.entity_payload.get('defaultInBuilding')

    @property
    def external_id(self):
        """External id

        Returns: External reference id of the provider

        """
        return self.entity_payload.get('externalId')

    @property
    def available(self):
        """Available

        Returns: True if door is currently available

        """
        return self.entity_payload.get('available')

    @property
    def order(self):
        """Order

        Returns: Integer that represents a linear order within the building

        """
        return self.entity_payload.get('order')

    def open(self):
        """Unlock the door

        """
        url = C_API_URI + C_DOOR_OPEN_ENDPOINT.format(self.entity_id)
        self._api.authenticated_query(url, method='post')



# coding: utf-8
"""Carson Living Authentication Module"""

import logging
import time
import requests
import jwt
from jwt import InvalidTokenError


from carson_living.const import (BASE_HEADERS,
                                 C_API_URI,
                                 C_AUTH_ENDPOINT)
from carson_living.util import default_carson_response_handler
from carson_living.error import (CarsonAPIError,
                                 CarsonAuthenticationError,
                                 CarsonTokenError)

_LOGGER = logging.getLogger(__name__)


# pylint: disable=useless-object-inheritance
class CarsonAuth(object):
    """A generalized Authentication Class for Carson Living.

    Responsible for managing (retrieving and updating) the
    JWT Authentication token for the Carson API.

    Attributes:
        _username: Carson Living username
        _password: Carson Living password
        _token: current JWT token
        _token_payload: current JWT token payload
        _token_expiration_time: current JWT token expiration time
        _token_update_cb:
            gets executed whenever the token gets update to a
            non-None value.
    """

    def __init__(self, username, password,
                 initial_token=None, token_update_cb=None):
        self._username = username
        self._password = password
        self._token = None
        self._token_payload = None
        self._token_expiration_time = None
        self._token_update_cb = None

        # Set and init token values
        self.token = initial_token

        # Set token updater after initial token (so it does not fire)
        self._token_update_cb = token_update_cb

    @property
    def username(self):
        """Username

        Returns:
            the configured username

        """
        return self._username

    @property
    def token(self):
        """
        Returns:
            current JWT token or None if currently authenticated.
        """
        return self._token

    @property
    def token_payload(self):
        """
        Returns:
            current JWT token payload or None if currently authenticated.
        """
        return self._token_payload

    @property
    def token_expiration_date(self):
        """
        Returns:
            current JWT token expiration time (seconds from epoch)
            or None if currently authenticated.
        """
        return self._token_expiration_time

    @token.setter
    def token(self, token):
        """Set or clear a new JWT Token.
        Args:
            token: Valid JWT token or None to clear current token.

        Raises:
            CarsonTokenError: JWT token format is invalid.
        """
        if token is None:
            self._token = None
            self._token_payload = None
            self._token_expiration_time = None
            return
        try:
            self._token_payload = jwt.decode(token, verify=False)
            self._token_expiration_time = self._token_payload.get('exp')

            self._token = token

            if self._token_update_cb is not None:
                self._token_update_cb(token)
            _LOGGER.info('Set access Token for %s',
                         self._token_payload.get('email', '<no e-mail found>'))
        except InvalidTokenError:
            raise CarsonTokenError('Cannot decode invalid token {}'
                                   .format(token))

    def update_token(self):
        """Authenticate user against Carson Living API.

        Raises:
            CarsonAuthenticationError: On authentication error.

        """
        _LOGGER.info('Getting new access token for %s', self._username)

        response = requests.post(
            (C_API_URI + C_AUTH_ENDPOINT),
            json={
                'username': self._username,
                'password': self._password,
            },
            headers=BASE_HEADERS
        )
        try:
            data = default_carson_response_handler(response)
            self.token = data.get('token')
            return self.token
        except CarsonAPIError as error:
            _LOGGER.warning('Authentication for %s failed', self._username)
            raise CarsonAuthenticationError(error)

    def valid_token(self):
        """Checks that Carson Authentication has a valid token.

        Returns:
            True if a token is set and not expired, otherwise False
        """
        if self.token is None or self._token_expiration_time is None:
            return False

        return self._token_expiration_time > int(time.time())

    def authenticated_query(self, url, method='get', params=None,
                            json=None, retry_auth=1,
                            response_handler=default_carson_response_handler):
        """Perform an authenticated Query against Carson Living

        Args:
            url: the url to query
            method: the http method to use
            params: the http params to use
            json: the json payload to submit
            retry_auth: number of query and reauthentication retries
            response_handler: dynamic response handler for api

        Returns:
            The unwrapped data dict of the Carson Living response.

        Raises:
            CarsonCommunicationError: Response was not received or
                not in the expected format.
            CarsonAPIError: Response indicated an client-side API
                error.
        """

        if not self.valid_token():
            self.update_token()

        headers = {'Authorization': 'JWT {}'.format(self.token)}
        headers.update(BASE_HEADERS)

        response = requests.request(method, url,
                                    headers=headers,
                                    params=params,
                                    json=json)

        # special case, clear token and retry. (Recursion)
        if response.status_code == 401 and retry_auth > 0:
            self.token = None
            return self.authenticated_query(
                url, method, params, json, retry_auth - 1,
                response_handler)

        return response_handler(response)



# -*- coding: utf-8 -*-
"""Collection of util functions"""

import time

from carson_living.error import (CarsonAPIError,
                                 CarsonCommunicationError)
from carson_living.const import CARSON_RESPONSE


def default_carson_response_handler(response):
    """Safely handle Carson API responses

    Args:
        response: A Python Requests response object.

    Returns:
        The unwrapped data dict of the Carson Living response.

    Raises:
        CarsonCommunicationError: Response was not received or
            not in the expected format.
        CarsonAPIError: Response indicated an client-side API
            error.
    """
    try:
        r_json = response.json()
        if not all(k in r_json for k in CARSON_RESPONSE.values()):
            raise CarsonCommunicationError(
                'Carson API response does not contain all expected keys')

        if r_json.get(CARSON_RESPONSE['CODE']) != 0:
            raise CarsonAPIError(
                # pylint: disable=too-many-format-args
                'Carson API error returned unsuccessful state. '
                'Status: {}, Message: {}'.format(
                    r_json.get(CARSON_RESPONSE['STATUS'], '<no status>'),
                    r_json.get(CARSON_RESPONSE['MSG']), '<no msg>')
                )
    except ValueError:
        raise CarsonCommunicationError(
            'Unable to handle response payload for {} to {}'.format(
                response.request.method, response.url))

    return r_json.get(CARSON_RESPONSE['DATA'])


def update_dictionary(current_dict, update_dict, constructor):
    """Update current_dict to update_dict without reconstructing existing

    update_dictionary updates the dict current_dict to resemble update_dict
    by
        1. removing values from current_dict that are missing in update_dict.
           (via del current_dict('missing'))
        2. adding values to current_dict that are only found in update_dict.
           (via constructor(update_dict['new'])
        3. update values (via current_dict[i].update(update_dict['changed']))

    Args:
        current_dict: The dict to update with entities
        update_dict: The latest dict with update payloads
        constructor: Constructor funtion to generate entity with payload

    """

    existing_keys = set(current_dict.keys())
    update_keys = set(update_dict.keys())

    # Update
    for i in existing_keys.intersection(update_keys):
        current_dict[i].update(update_dict[i])

    # Add
    for i in update_keys.difference(existing_keys):
        current_dict[i] = constructor(update_dict[i])

    # Remove
    for i in existing_keys.difference(update_keys):
        del current_dict[i]


def current_milli_time():
    """Return the current time in milliseconds"""
    return int(time.time() * 1000)


def timedelta_to_milli_time(timedelta):
    """Return the current time in milliseconds"""
    return int(timedelta.total_seconds() * 1000)



"""Basic Eagle Eye API Module"""
import logging

import requests
from requests import HTTPError

from carson_living.error import (CarsonError,
                                 CarsonAPIError)
from carson_living.eagleeye_entities import EagleEyeCamera

from carson_living.util import update_dictionary
from carson_living.const import (BASE_HEADERS,
                                 EEN_API_URI,
                                 EEN_DEVICE_LIST_ENDPOINT,
                                 EEN_IS_AUTH_ENDPOINT)

_LOGGER = logging.getLogger(__name__)


# pylint: disable=useless-object-inheritance
class EagleEye(object):
    """Eagle Eye API class for interfacing with the endpoints

    This class should probably be moved in a dedicated Eagle Eye project,
    but initially it can live within Carson Living. Note, the eagle eye
    API does not update it's state during initialization, but is updated
    externally. Carson Living update automatically triggers an update call
    to Eagle Eye.
    """

    def __init__(self, session_callback):
        self._session_callback = session_callback
        self._session_auth_key = None
        self._session_brand_subdomain = None
        self._cameras = {}

    @property
    def session_auth_key(self):
        """Current Auth Key"""
        return self._session_auth_key

    @property
    def session_brand_subdomain(self):
        """Current Brand Subdomain"""
        return self._session_brand_subdomain

    @property
    def cameras(self):
        """Get all cameras returned directly by the API"""
        return self._cameras.values()

    def get_camera(self, ee_id):
        """

        Args:
            ee_id: Eagle Eye camera id

        Returns:
            The EagleEye Camera with id or None, if not found.

        """
        return self._cameras.get(ee_id)

    def update_session_auth_key(self):
        """Updates the internal session state via session_callback

        Raises:
            CarsonError: If callback returns empty value.

        """
        _LOGGER.debug(
            'Trying to update the session auth key for the Eagle Eye API.')
        auth_key, brand_subdomain = self._session_callback()

        if not auth_key or not brand_subdomain:
            raise CarsonError(
                'Eagle Eye authentication callback returned empty values.')

        self._session_auth_key = auth_key
        self._session_brand_subdomain = brand_subdomain

    def check_auth(self, refresh=True):
        """Check if the current auth_key is still valid

        Args:
            refresh:
                automatically update auth_key if not valid

        Returns: True if a valid auth_key exists.

        """
        if not refresh and not self._session_auth_key:
            return False

        retry_auth = 1 if refresh else 0

        try:
            self.authenticated_query(
                EEN_API_URI + EEN_IS_AUTH_ENDPOINT,
                retry_auth=retry_auth
            )
        except CarsonAPIError:
            return False

        return True

    def authenticated_query(self, url, method='get', params=None,
                            json=None, retry_auth=1, stream=None,
                            response_handler=lambda r: r.json()):
        """Perform an authenticated Query against Eagle Eye

        Args:
            url:
                the url to query, can contain a branded subdomain
                to substitute
            method: the http method to use
            params: the http params to use
            json: the json payload to submit
            retry_auth: number of query and reauthentication retries
            stream: Stream the content
            response_handler: optional file handler to stream the raw content

        Returns:
            The json response object, or the file handler that was passed
            to receive the raw response.

        Raises:
            CarsonAPIError: Response indicated an client or
            server-side API error.
        """
        if not self._session_auth_key \
                or not self._session_brand_subdomain:
            self.update_session_auth_key()

        headers = {'Cookie': 'auth_key={}'.format(self._session_auth_key)}
        headers.update(BASE_HEADERS)

        response = requests.request(method,
                                    url.format(self._session_brand_subdomain),
                                    headers=headers,
                                    params=params,
                                    json=json,
                                    stream=stream)

        # special case, clear token and retry. (Recursion)
        if response.status_code == 401 and retry_auth > 0:
            _LOGGER.info(
                'Eagle Eye request %s returned 401, retrying ... (%d left)',
                url, retry_auth)
            self._session_auth_key = None
            return self.authenticated_query(
                url, method, params, json, retry_auth - 1,
                stream, response_handler)

        try:
            response.raise_for_status()
            return response_handler(response)

        except HTTPError as error:
            raise CarsonAPIError(error)

    def update(self):
        """Update internal state

        Update entity list and individual entity parameters associated with the
        Eagle Eye API

        """
        _LOGGER.debug('Updating Eagle Eye API and associated entities')
        self._update_cameras()

    def _update_cameras(self):
        # Query List
        device_list = self.authenticated_query(
            EEN_API_URI + EEN_DEVICE_LIST_ENDPOINT
        )

        update_cameras = {
            c[1]: EagleEyeCamera.map_list_to_entity_payload(c)
            for c in device_list if c[3] == 'camera'
        }

        update_dictionary(
            self._cameras,
            update_cameras,
            lambda c: EagleEyeCamera(self, c))



# -*- coding: utf-8 -*-
"""Python Module to use with Carson Living

Contains classes to authenticate against Carson Living and query various
entities and expose their functionality.
"""

from carson_living.auth import CarsonAuth
from carson_living.carson import Carson
from carson_living.error import (CarsonAuthenticationError,
                                 CarsonAPIError,
                                 CarsonError,
                                 CarsonCommunicationError,
                                 CarsonTokenError)

from carson_living.eagleeye import EagleEye
from carson_living.eagleeye_entities import EagleEyeCamera
from carson_living.carson_entities import (CarsonDoor,
                                           CarsonBuilding,
                                           CarsonUser)

from carson_living.const import (EEN_ASSET_REF_ASSET,
                                 EEN_ASSET_REF_PREV,
                                 EEN_ASSET_REF_NEXT,
                                 EEN_ASSET_REF_AFTER,
                                 EEN_ASSET_CLS_ALL,
                                 EEN_ASSET_CLS_PRE,
                                 EEN_ASSET_CLS_THUMB,
                                 EEN_VIDEO_FORMAT_FLV,
                                 EEN_VIDEO_FORMAT_MP4)


__all__ = ['CarsonAuth',
           'Carson',
           'CarsonAuthenticationError',
           'CarsonAPIError',
           'CarsonError',
           'CarsonCommunicationError',
           'CarsonTokenError',
           'EagleEye',
           'EagleEyeCamera',
           'CarsonDoor',
           'CarsonBuilding',
           'CarsonUser',
           'EEN_ASSET_REF_ASSET',
           'EEN_ASSET_REF_PREV',
           'EEN_ASSET_REF_NEXT',
           'EEN_ASSET_REF_AFTER',
           'EEN_ASSET_CLS_ALL',
           'EEN_ASSET_CLS_PRE',
           'EEN_ASSET_CLS_THUMB',
           'EEN_VIDEO_FORMAT_FLV',
           'EEN_VIDEO_FORMAT_MP4']



# -*- coding: utf-8 -*-
"""Carson Living custom exceptions."""


class CarsonError(Exception):
    """Carson Living base error"""


class CarsonTokenError(CarsonError):
    """Carson Living token error"""


class CarsonCommunicationError(CarsonError):
    """Carson Living communication error"""


class CarsonAPIError(CarsonError):
    """Carson Living client-side API error"""


class CarsonAuthenticationError(CarsonAPIError):
    """Carson Living authentication error"""



# coding=utf-8
"""Python Carson Living setup script."""
from setuptools import setup

_VERSION = '0.0.4'


def readme():
    """Pipe README.rst"""
    with open('README.rst') as desc:
        return desc.read()


setup(
    name='carson_living',
    packages=['carson_living'],
    version=_VERSION,
    description='A Python library to communicate with'
                ' Carson Living Residences (https://www.carson.live/)',
    long_description=readme(),
    author='Martin Riedel',
    author_email='web@riedel-it.de',
    url='https://github.com/rado0x54/python-carson-living',
    license='Apache License 2.0',
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
    include_package_data=True,
    install_requires=['requests', 'pyjwt'],
    test_suite='tests',
    keywords=[
        'carson living',
        'virtual doorman',
        'home automation',
        ],
    classifiers=[
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Home Automation',
        'Topic :: Software Development :: Libraries :: Python Modules'
        ],
)



# -*- coding: utf-8 -*-
"""Carson Living API Module."""
import logging

from carson_living.auth import CarsonAuth

from carson_living.carson_entities import (CarsonUser,
                                           CarsonBuilding)
from carson_living.const import (C_API_URI,
                                 C_ME_ENDPOINT)
from carson_living.util import update_dictionary


_LOGGER = logging.getLogger(__name__)


# pylint: disable=useless-object-inheritance
class Carson(CarsonAuth):
    """A Python Abstraction object to the Carson Living API.

        Attributes:
            _user: The authenticated user to Carson
            _buildings:
                The building properties that are associated with
                the current user
    """
    def __init__(self, username, password,
                 initial_token=None, token_update_cb=None):
        super(Carson, self).__init__(username, password,
                                     initial_token, token_update_cb)

        self._user = None
        self._buildings = {}

        self.update()

    @property
    def buildings(self):
        """Building properties that belong to the user"""
        return self._buildings.values()

    @property
    def first_building(self):
        """Convenience Function to return first building in account"""
        return next(iter(self.buildings))

    @property
    def user(self):
        """The current authenticated user"""
        return self._user

    def update(self):
        """Update entity list and individual entity parameters associated with the API

        """
        _LOGGER.debug('Updating Carson Living API and associated entities')
        url = C_API_URI + C_ME_ENDPOINT
        me_payload = self.authenticated_query(url)

        self._update_user(me_payload)
        self._update_buildings(me_payload)

    def _update_user(self, payload):
        if self._user is None:
            self._user = CarsonUser(entity_payload=payload)
        else:
            self._user.update(payload)

    def _update_buildings(self, payload):
        # Not 100% if propertyLevel condition is playing it overly safe.
        update_buildings = {p['id']: p
                            for p in payload.get('properties')
                            if p['propertyLevel'] == 'building'}

        update_dictionary(
            self._buildings,
            update_buildings,
            lambda p: CarsonBuilding(
                self,
                p))



# coding: utf-8
"""Carson Constants."""

CARSON_RESPONSE = {
    'CODE': 'code',
    'STATUS': 'status',
    'DATA': 'data',
    'MSG': 'msg'
}

"""Pretend to be Carson iOS Installation v1.0.171"""
BASE_HEADERS = {
    'User-Agent': 'Carson/1.0.200 (live.carson.app; build:315; iOS 13.3.0) '
                  'Alamofire/1.0.200',
    'X-App-Version': '1.0.200(315)',
    'X-Device-Type': 'ios'
}

# number of attempts to refresh token
# RETRY_TOKEN = 3

# default suffix for session cache file
# CACHE_ATTRS = {'account': None, 'alerts': None, 'token': None}
#
# try:
#     CACHE_FILE = os.path.join(os.getenv("HOME"),
#                               '.carson_living-session.cache')
# except (AttributeError, TypeError):
#     CACHE_FILE = os.path.join('.', '.carson_living-session.cache')


# code when item was not found
# NOT_FOUND = -1

# Carson API endpoints
# Beware URLs end in '/', otherwise it returns a
# HTTP/1.1 301 Moved Permanently to the correct version.
C_API_VERSION = 'v1.4.1'
C_API_URI = 'https://api.carson.live/api/' + C_API_VERSION

C_AUTH_ENDPOINT = '/auth/login/'
C_ME_ENDPOINT = '/me/'

C_DOOR_OPEN_ENDPOINT = '/doors/{}/open/'
C_EEN_SESSION_ENDPOINT = '/properties/buildings/{}/eagleeye/session/'

# Eagle Eye API endpoints
# Beware URLs DO NOT end in '/', otherwise it returns a 500
EEN_API_URI = 'https://{}.eagleeyenetworks.com'
EEN_DEVICE_ENDPOINT = '/g/device'
EEN_DEVICE_LIST_ENDPOINT = '/g/device/list'
EEN_GET_IMAGE_ENDPOINT = '/asset/{}/image.jpeg'
EEN_GET_VIDEO_ENDPOINT = '/asset/play/video.{}'
EEN_IS_AUTH_ENDPOINT = '/g/aaa/isauth'

# Eagle Eye Network Interface options
EEN_ASSET_REF_ASSET = 'asset'
EEN_ASSET_REF_PREV = 'prev'
EEN_ASSET_REF_NEXT = 'next'
EEN_ASSET_REF_AFTER = 'after'

EEN_ASSET_CLS_ALL = 'all'
EEN_ASSET_CLS_PRE = 'pre'
EEN_ASSET_CLS_THUMB = 'thumb'

EEN_VIDEO_FORMAT_FLV = 'flv'
EEN_VIDEO_FORMAT_MP4 = 'mp4'



"""Eagle Eye API Entities"""
import shutil

from requests import Request

from carson_living.entities import _AbstractAPIEntity

from carson_living.const import (EEN_API_URI,
                                 EEN_DEVICE_ENDPOINT,
                                 EEN_GET_IMAGE_ENDPOINT,
                                 EEN_GET_VIDEO_ENDPOINT,
                                 EEN_ASSET_CLS_PRE,
                                 EEN_ASSET_REF_PREV,
                                 EEN_VIDEO_FORMAT_FLV)

from carson_living.error import CarsonAPIError

from carson_living.util import (current_milli_time,
                                timedelta_to_milli_time)


class EagleEyeCamera(_AbstractAPIEntity):
    """Eagle Eye Camera Entity

    The eagle eye camera is initialized with the device/list payload to
    allow for fast initialization

    """
    def __init__(self, api, entity_payload):
        super(EagleEyeCamera, self).__init__(
            api,
            update_callback=self._get_payload_internal,
            entity_payload=entity_payload
        )

    @classmethod
    def from_api(cls, api, camera_id):
        """Init Camera from API call

        Args:
            api: Eagle Eye API
            camera_id: Eagle Eye Camera ID

        Returns:
            Initialized EagleEyeCamera

        """
        entity_payload = cls.get_payload(api, camera_id)
        return cls(api, entity_payload)

    @classmethod
    def from_list_payload(cls, api, list_entity_payload):
        """Init Camera from List Payload

        Args:
            api: Eagle Eye API
            list_entity_payload:
                Eagle Eye JSON from /list call

        Returns:
            Initialized EagleEyeCamera

        """
        entity_payload = cls.map_list_to_entity_payload(list_entity_payload)
        return cls(api, entity_payload)

    @staticmethod
    def map_list_to_entity_payload(list_entity_payload):
        """Map from list to entity payload

        Args:
            list_entity_payload:
                Eagle eye list payload

        Returns:
            Eagle eye entity payload

        """
        return {
            "bridges": {
                b[0]: b[1] for b in list_entity_payload[4]
            },
            "name": list_entity_payload[2],
            "tags": list_entity_payload[7],
            "utcOffset": list_entity_payload[12],
            "timezone": list_entity_payload[11],
            "permissions": list_entity_payload[6],
            "guid": list_entity_payload[8],
            "id": list_entity_payload[1],
            "account_id": list_entity_payload[0]
        }

    def _get_payload_internal(self):
        return self.get_payload(self._api, self.entity_id)

    @staticmethod
    def get_payload(api, camera_id):
        """Get entity payload from API

        Args:
            api: Eagle Eye API
            camera_id: Eagle Eye Camera ID

        Returns:
            Eagle eye entity payload

        """
        url = EEN_API_URI + EEN_DEVICE_ENDPOINT
        return api.authenticated_query(
            url, params={'id': camera_id})

    @property
    def unique_entity_id(self):
        return 'eagleeye_camera_{}'.format(self.entity_id)

    def _internal_update(self):
        pass

    @property
    def entity_id(self):
        return self._entity_payload.get('id')

    def __str__(self):
        pattern = """\
id: {entity_id}
name: {name}
account id: {account_id}
guid: {guid}
tags: {tags}"""
        return pattern.format(
            entity_id=self.entity_id,
            name=self.name,
            account_id=self.account_id,
            guid=self.guid,
            tags=', '.join(self.tags)
        )

    @property
    def account_id(self):
        """Account id

        Returns: Eagle Eye Account Id for that camera

        """
        return self._entity_payload.get('account_id')

    @property
    def name(self):
        """Name

        Returns: Device name

        """
        return self._entity_payload.get('name')

    @property
    def utc_offset(self):
        """UTC offset

        Returns: Signed UTC offset in seconds of the set 'timezone'

        """
        return self._entity_payload.get('utcOffset')

    @property
    def timezone(self):
        """Timezone

        Returns: tz database string of the camera

        """
        return self._entity_payload.get('timezone')

    @property
    def guid(self):
        """GUID

        Returns:
            The GUID (Globally Unique Identifier) is an immutable device
            identifier

        """
        return self._entity_payload.get('guid')

    @property
    def permissions(self):
        """Permissions

        Returns:
            String of characters each defining a permission level of
            the current user

        """
        return self._entity_payload.get('permissions')

    @property
    def tags(self):
        """Tags

        Returns:
            Array of strings each representing a tag name

        """
        return self._entity_payload.get('tags')

    @property
    def bridges(self):
        """Bridges

        Returns:
            Json object of bridges (ESNs) this device is seen by and the
            camera attach status:

        """
        return self._entity_payload.get('bridges')

    @staticmethod
    def utc_to_een_timestamp(utc_dt):
        """Convert utc_dt to EEN string

        Note, no timezone conversion is performed, please make sure the
        Datetime object is in UTC.

        Args:
            utc_dt: Datetime object in UTC

        Returns: EEN timestamp format

        """
        return utc_dt.strftime('%Y%m%d%H%M%S.%f')[:-3]

    @staticmethod
    def _get_video_timestamps(length, utc_dt, video_format):
        # default are download parameters
        time_millies = current_milli_time()
        length_millies = timedelta_to_milli_time(length)

        # Live case
        start_ts = 'stream_{}'.format(time_millies)
        end_ts = '+{}'.format(length_millies)

        if utc_dt is None:
            if video_format != EEN_VIDEO_FORMAT_FLV:
                raise CarsonAPIError(
                    'Live video streaming is only possible with .flv')
        else:
            # Not live
            start_ts = EagleEyeCamera.utc_to_een_timestamp(utc_dt)
            end_ts = EagleEyeCamera.utc_to_een_timestamp(utc_dt + length)

        return start_ts, end_ts

    def get_image(self, file,
                  utc_dt=None,
                  asset_ref=EEN_ASSET_REF_PREV,
                  asset_class=EEN_ASSET_CLS_PRE):
        """Get binary JPEG image from the camera

        Args:
            file:
                file handler that is written to.
            utc_dt:
                Datetime object in UTC
            asset_ref:
                prev: previous image to time stamp
                next: next image to timestamp (blocks)
                asset: image at timestamp
            asset_class:
                all, pre, thumb

        Returns:
            JPEG Image
        """
        def _response_file_handler(response):
            response.raw.decode_content = True
            shutil.copyfileobj(response.raw, file)

        timestamp = 'now'
        if utc_dt is not None:
            timestamp = self.utc_to_een_timestamp(utc_dt)

        url = EEN_API_URI + EEN_GET_IMAGE_ENDPOINT.format(
            asset_ref)
        return self._api.authenticated_query(
            url, params={'id': self.entity_id,
                         'timestamp': timestamp,
                         'asset_class': asset_class},
            stream=True,
            response_handler=_response_file_handler)

    # Not this is quite duplicate at the moment, but a major refactor
    # would be needed to return a prepared url via authenticated_query
    def get_image_url(self, utc_dt=None,
                      asset_ref=EEN_ASSET_REF_PREV,
                      asset_class=EEN_ASSET_CLS_PRE,
                      check_auth=True):
        """Get JPEG image URL from the camera

        Args:
            utc_dt:
                Datetime object in UTC
            asset_ref:
                prev: previous image to time stamp
                next: next image to timestamp (blocks)
                asset: image at timestamp
            asset_class:
                all, pre, thumb
            check_auth:
                Check auth token and refresh

        Returns:
            JPEG Image URL or None if not valid Auth exists
        """
        if check_auth and not self._api.check_auth():
            return None

        timestamp = 'now'
        if utc_dt is not None:
            timestamp = self.utc_to_een_timestamp(utc_dt)

        url = EEN_API_URI.format(
            self._api.session_brand_subdomain
        ) + EEN_GET_IMAGE_ENDPOINT.format(
            asset_ref
        )

        prepared = Request(url=url, params={
            'id': self.entity_id,
            'timestamp': timestamp,
            'asset_class': asset_class,
            'A': self._api.session_auth_key}).prepare()
        return prepared.url

    # stream Live video to file
    def get_video(self, file, length, utc_dt=None,
                  video_format=EEN_VIDEO_FORMAT_FLV):
        """Get a (live) video stream from the camera

        Args:
            file: file handler for the response
            length: of the stream in timedelta
            video_format: flv or mp4
            utc_dt: utc timestamp for video, live for None

        Returns:
            Video stream to file
        """
        def _response_file_handler(response):
            response.raw.decode_content = True
            shutil.copyfileobj(response.raw, file)

        start_ts, end_ts = self._get_video_timestamps(
            length, utc_dt, video_format)

        url = EEN_API_URI + EEN_GET_VIDEO_ENDPOINT.format(
            video_format)
        return self._api.authenticated_query(
            url, params={'id': self.entity_id,
                         'start_timestamp': start_ts,
                         'end_timestamp': end_ts},
            stream=True,
            response_handler=_response_file_handler)

    # Not this is quite duplicate at the moment, but a major refactor
    # would be needed to return a prepared url via authenticated_query
    def get_video_url(self, length, utc_dt=None,
                      video_format=EEN_VIDEO_FORMAT_FLV, check_auth=True):
        """Get a (live) video stream from the camera

        Args:
            file: file handler for the response
            length: of the stream in timedelta
            video_format: flv or mp4
            utc_dt: utc timestamp for video, live for None
            check_auth: Check auth token and refresh

        Returns:
            Video url or None if not valid Auth exists
        """
        if check_auth and not self._api.check_auth():
            return None

        start_ts, end_ts = self._get_video_timestamps(
            length, utc_dt, video_format)

        url = EEN_API_URI.format(
            self._api.session_brand_subdomain
        ) + EEN_GET_VIDEO_ENDPOINT.format(
            video_format
        )

        prepared = Request(url=url, params={
            'id': self.entity_id,
            'start_timestamp': start_ts,
            'end_timestamp': end_ts,
            'A': self._api.session_auth_key}).prepare()
        return prepared.url
